// Copyright 2021 Northern.tech AS
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

package main

import (
	"fmt"
	"os"

	"github.com/mendersoftware/mender-artifact/cli"
)

func run() error {
	err := cli.Run(os.Args)
	if err != nil {
		fmt.Println(err)
	}
	return err
}

func main() {
	if err := run(); err != nil {
		os.Exit(1)
	}
}
